from typing import List
from queue import Queue


class Message:
    pass

class ClientDoneMessage:
    def __init__(self, id: int):
        self.id = id

class MessageBus:

    _messages = Queue()

    @classmethod
    def put(cls, message: Message):
        cls._messages.put(message)

    @classmethod
    def get(cls):
        if not cls._messages.empty():
            return cls._messages.get_nowait()
        return None



from datetime import datetime

from avss.http import HttpRequest, HttpResponse


class HttpParser:

    DEFAULT_VERSION = 'HTTP/1.0'
    DEFAULT_HEADERS = {
        'Content-Type': 'text/html; charset=utf-8',
        'Server': 'avss'
    }

    def parse(data: str):
        try:
            lines = data.splitlines()
            method, path, version = lines[0].split(' ')

            headers = {}
            body_index = None
            for index, line in enumerate(lines[1:]):
                if line == '\r\n':
                    body_index = index
                    break
                if line:
                    key, val = line.split(':', 1)
                    headers[key] = val

            if body_index is not None:
                body = lines[body_index:]
            else:
                body = ''

            return HttpRequest(headers, body, method, path, version)
        except IndexError:
            print(f'ERR {data}')

    @classmethod
    def get_headers(cls, kwargs: dict, data: str):
        headers = dict(cls.DEFAULT_HEADERS)
        headers['Date'] = datetime.now().strftime('%Y-%m-%d %H:%M:%S')
        headers['Content-Length'] = len(data)
        headers.update(kwargs)
        return headers

    @classmethod
    def make_response(cls, data: str, **kwargs):

        response = HttpResponse(
            version=kwargs.get('version', cls.DEFAULT_VERSION),
            status=kwargs.get('status', 200),
            phrase=kwargs.get('phrase', 'OK'),
            headers=cls.get_headers(kwargs, data),
            body=data
        )

        return response



from abc import ABC, abstractclassmethod
import socket
import threading
from typing import List
from queue import Queue

from avss.connection import Connection, HttpConnection
from avss.settings import get_logger, get_settings
from avss.message_bus import MessageBus, Message, ClientDoneMessage


logger = get_logger()
settings = get_settings()


class Server(ABC):

    def __init__(self, host: str, port: int):
        self.host = host
        self.port = port
        self.max_connections = int(settings['settings']['MAX_CONNECTIONS'])

        self._connections: Dict[int, Connection] = {}
        self._pending_connections = Queue()
        self._running = threading.Event()

    @abstractclassmethod
    def do_start(self):
        pass

    @abstractclassmethod
    def do_run(self):
        pass

    @abstractclassmethod
    def do_stop(self):
        pass

    def check_message_bus(self):
        message = MessageBus.get()
        if message is not None:
            if isinstance(message, ClientDoneMessage):
                logger.debug(f'Client {message.id} is done, removing reference con.')
                self._connections.pop(message.id)

    def respond_client(self, con: Connection):
        if len(self._connections) < self.max_connections:
            con.start()
            self._connections[con.id] = con
        else:
            self._pending_connections.put(con)

    def start(self):
        self.do_start()
        self._running.set()
        self.do_run()

    def stop(self):
        self._running.clear()
        self.do_stop()

    def is_running(self):
        return self._running.is_set()


class HttpServer(Server):

    def __init__(self, host='', port=6001):
        super().__init__(host, port)
        self._sock = None
        self._listen_timout = 1

    def do_start(self):
        self._sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self._sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        self._sock.settimeout(self._listen_timout)
        self._sock.bind((self.host, self.port))
        logger.info(f'Server listening at {self.host}:{self.port}')

    def do_run(self):
        if self._sock is None:
            logging.error('Cant run server without socket connection.')
            return

        self._sock.listen(0)

        while self.is_running():
            try:
                conn, addr = self._sock.accept()
                logger.debug(f'New connection at {addr}')
                self.respond_client(HttpConnection(conn, addr))
            except socket.timeout:
                self.check_message_bus()

    def do_stop(self):
        pass



from abc import ABC, abstractclassmethod

from avss.http import HttpRequest, HttpResponse
from avss.parser import HttpParser
from avss.settings import get_settings
import importlib
import sys


settings = get_settings()

class AppFactory:

    @classmethod
    def get_app(cls, http_request: HttpRequest):
        # TODO: Cache these on start.
        app = None
        for section in settings.sections():
            if section not in ['settings', 'default']:
                app_path = settings[section].get('path')
                app_name = settings[section].get('app')
                sys.path.append(app_path)
                app = importlib.import_module(app_name)
                app = getattr(app, 'main')
                print(app)

        if app is None:
            from avss.app import DefaultApp
            return DefaultApp

        return app




from abc import ABC, abstractclassmethod
from dataclasses import dataclass, field
from typing import Dict


@dataclass
class HttpPacket(ABC):
    headers: Dict[str, str]
    body: str

    def as_string(self):
        packet = self.get_header() + '\r\n'
        packet += '\r\n'.join(f'{key}: {val}' for key, val in self.headers.items())
        packet += '\r\n\r\n'
        if self.body:
            packet += self.body
            packet += '\r\n\r\n'

        return packet

    def as_bytes(self):
        return self.as_string().encode('utf-8')

    @abstractclassmethod
    def get_header(self):
        pass


@dataclass
class HttpRequest(HttpPacket):
    method: str
    path: str
    version: str

    def get_header(self):
        return f'{self.method} {self.path} {self.version}'


@dataclass
class HttpResponse(HttpPacket):
    version: str
    status: str
    phrase: str

    def get_header(self):
        return f'{self.version} {self.status} {self.phrase}'



from abc import ABC, abstractclassmethod
import threading

from avss.app_factory import AppFactory
from avss.message_bus import MessageBus, Message, ClientDoneMessage
from avss.settings import get_logger
from avss.parser import HttpParser

logger = get_logger()


class Connection(threading.Thread, ABC):

    def __init__(self, con, addr):
        self.con = con
        self.addr = addr
        self.id = addr[1]
        self._running = threading.Event()
        super().__init__(daemon=True)

    def run(self):
        self._running.set()
        self.do_run()

    def close(self):
        self._running.clear()
        self.con.close()
        MessageBus.put(ClientDoneMessage(self.id))

    def is_running(self):
        return self._running.is_set()

    def __hash__(self):
        return hash(self.id)

    def __eq__(self, other):
        return isinstance(other, self.__class__) and other.id == self.id

    @abstractclassmethod
    def do_run(self):
        pass


class HttpConnection(Connection):

    def do_run(self):
        while self.is_running():
            # Receive message & parse Http Request
            msg = self.con.recv(2048)
            http_request = HttpParser.parse(msg.decode())

            # Find the correct app to serve with the given request.
            app = AppFactory.get_app(http_request)
            http_response = app(http_request)

            # Send the response to client.
            self.con.send(http_response.as_bytes())
            logger.debug(http_response.as_string())
            self.close()


class HttpsConnection(Connection):

    def do_run(self):
        pass


from avss.http import HttpRequest, HttpResponse
from avss.parser import HttpParser


def default_main(self, http_request: HttpRequest) -> HttpResponse:
    return HttpParser.make_response('<h1>Hello World</h1>')



import configparser
import logging
from pathlib import Path

global _LOGGER
_LOGGER = None
global _SETTINGS
_SETTINGS = None

BASE_PATH = Path(__file__).parent


def get_settings():
    global _SETTINGS
    if _SETTINGS is None:
        _SETTINGS = configparser.ConfigParser()
        with open(BASE_PATH.joinpath('config.ini')) as f:
            _SETTINGS.read_file(f)
    return _SETTINGS


def get_logger():
    global _LOGGER
    if _LOGGER is None:
        logging.basicConfig()
        logging.root.setLevel(logging.DEBUG)
        _LOGGER = logging.getLogger()

    return _LOGGER
